#include "Windows/WavAudioRender.h"

#include <windows.h>
#include <mmsystem.h>

static WAVEHDR* AllocateBlocks(uint32 size, int32 count) {
    uint32 totalBufferSize = (size + sizeof(WAVEHDR)) * count;

    uint8* buffer{ nullptr };
    if ((buffer = reinterpret_cast<uint8*>(HeapAlloc(
        GetProcessHeap(),
        HEAP_ZERO_MEMORY,
        totalBufferSize
    ))) == NULL) {
        ExitProcess(1);
    }

    WAVEHDR* blocks = reinterpret_cast<WAVEHDR*>(buffer);
    buffer += sizeof(WAVEHDR) * count;
    for (int32 index = 0; index < count; ++index) {
        blocks[index].dwBufferLength = size;
        blocks[index].lpData = (LPSTR)buffer;
        blocks[index].dwFlags = 0;
        buffer += size;
    }
    return blocks;
}

static void FreeBlocks(WAVEHDR* blockArray) {
    HeapFree(GetProcessHeap(), 0, blockArray);
}

WavAudioRender::WavAudioRender() noexcept {
    HWAVEOUT hWaveOut = nullptr;

    waveBlocks_ = AllocateBlocks(BlockSize_, BlockCount_);
    
    constexpr int32 rate = 16000;
    constexpr int32 channel = 1;
    constexpr int32 bitePerSample = 16;

    WAVEFORMATEX waveform;
    waveform.wFormatTag = WAVE_FORMAT_PCM;
    waveform.nChannels = channel;
    waveform.nSamplesPerSec = rate;
    waveform.nAvgBytesPerSec = rate * channel * 2;
    waveform.nBlockAlign = channel * 2;
    waveform.wBitsPerSample = bitePerSample;
    waveform.cbSize = 0;


    MMRESULT hr = waveOutOpen(&hWaveOut, WAVE_MAPPER, &waveform,
        reinterpret_cast<DWORD_PTR>(WavAudioRender::waveOutProc),
        reinterpret_cast<DWORD_PTR>(this), CALLBACK_FUNCTION);
    if (MMSYSERR_NOERROR == hr) {
        return;
    }

    hWavout_ = std::move(hWaveOut);
}

WavAudioRender::~WavAudioRender() {
    if (nullptr == hWavout_) {
        return;
    }
    waveOutReset(hWavout_);
    for (int i = 0; i < BlockCount_; i++) {
        if (waveBlocks_[i].dwFlags & WHDR_PREPARED) {
            waveOutUnprepareHeader(hWavout_, &waveBlocks_[i], sizeof(WAVEHDR));
        }
    }

    do {
        MMRESULT ret = waveOutClose(hWavout_);
        if (MMSYSERR_NOERROR == ret) {
            break;
        } else if (WAVERR_STILLPLAYING == ret) {
            ::Sleep(1);
        } else if (MMSYSERR_NOMEM == ret) {
            break;
        } else {
            break;
        }
    } while (true);
    FreeBlocks(waveBlocks_);
}

bool WavAudioRender::Write(const AudioFrame& audioFrame) {
    if (nullptr == hWavout_) {
        return false;
    }

    WAVEHDR* current;
    int32 remain;
    current = &waveBlocks_[waveCurrentBlock_];
    uint64 size = audioFrame.data_.size();
    const uint8* data = audioFrame.data_.data();
    while (size > 0) {
        if (current->dwFlags & WHDR_PREPARED) {
            waveOutUnprepareHeader(hWavout_, current, sizeof(WAVEHDR));
        }
        if (size < (int)(BlockSize_ - current->dwUser)) {
            memcpy(current->lpData + current->dwUser, data, size);
            current->dwUser += size;
            break;
        }
        remain = BlockSize_ - static_cast<int32>(current->dwUser);
        memcpy(current->lpData + current->dwUser, data, remain);
        size -= remain;
        data += remain;
        current->dwBufferLength = BlockSize_;
        waveOutPrepareHeader(hWavout_, current, sizeof(WAVEHDR));
        waveOutWrite(hWavout_, current, sizeof(WAVEHDR));
        {
            Locker<SectionLock> locker(lock_);
            --freeBlockCounter_;
        }

        while (!freeBlockCounter_) {
            Sleep(1);
        }

        ++waveCurrentBlock_;
        waveCurrentBlock_ %= BlockCount_;
        current = &waveBlocks_[waveCurrentBlock_];
        current->dwUser = 0;
    }
    return true;
}

void CALLBACK WavAudioRender::waveOutProc(HWAVEOUT hwo, UINT uMsg, DWORD_PTR dwInstance,
    DWORD_PTR dwParam1, DWORD_PTR dwParam2) {
    WavAudioRender* self = reinterpret_cast<WavAudioRender*>(dwInstance);

    if (uMsg != WOM_DONE)
        return;

    self->FreeBlock();
}

void WavAudioRender::FreeBlock() {
    Locker<SectionLock> locker(lock_);
    ++freeBlockCounter_;
}
